import numpy as np
import fastring

class RingBuffer:
    """
    Python adapter for C++ fastring class
    Internally the native ring is always (C_flat, S_flat) with S_flat=1 for images.
    """
    def __init__(self, 
                 name, 
                 capacity_frames, 
                 frame_shape, 
                 data_mode, 
                 dtype, 
                 create=False):
        self.name = name
        self.capacity = int(capacity_frames)
        self.logical_shape = tuple(frame_shape)
        self.dtype = np.dtype(dtype)
        self._mode = str(data_mode)

        if (self._mode == "line"):
            N, S, C = self.logical_shape
            self._N, self._S, self._C = int(N), int(S), int(C)
            self.frame_shape = (int(N), int(S), int(C))
        elif (self._mode == "image"):
            H, W, C = self.logical_shape
            self._H, self._W, self._Cimg = int(H), int(W), int(C)
            self._S = self._H * self._W * self._Cimg
            self.frame_shape = (self._H, self._W, self._Cimg)
        else:
            raise ValueError(f"frame_shape must be (N,S,C) or (H,W,C), got {self.logical_shape}")

        maker = fastring.Ring.create if create else fastring.Ring.open
        self._ring = maker(self.name, int(self.capacity), int(self._S * self.dtype.itemsize))

    @property
    def write_idx(self) -> int:
        return int(self._ring.write_idx)

    @property
    def frame_bytes(self) -> int:
        return int(self._ring.frame_bytes)

    def publish(self, arr):
        """
        Publish array to ring buffer
        :param arr: input array to store
        """
        a = np.asarray(arr)

        if self._mode == "line":
            if a.shape == (self._C, self._S):
                a = a.T
            if a.shape != (self._S, self._C):
                raise ValueError(f"publish LINE expects (S,C) got {a.shape}")
            frame = np.ascontiguousarray(a, dtype=self.dtype)
            self._ring.publish(frame)
            return
        else:
            # image mode
            if a.ndim == 3:
                if a.shape != (self._H, self._W, self._Cimg):
                    raise ValueError(f"publish Image expects (H,W,C), got {a.shape}")
                frame = np.ascontiguousarray(a, dtype=self.dtype)
                self._ring.publish(frame)
                return

            if a.ndim == 4 and a.shape[1:] == (self._H, self._W, self._Cimg):
                a = np.ascontiguousarray(a, dtype=self.dtype)
                for i in range(a.shape[0]):
                    self._ring.publish(a[i])
                return

            raise ValueError(f"publish Image expects (H,W,C) or (N,H,W,C), got {a.shape}")

    def view_window(self, start: int, frames: int):
        """
        Return a NumPy view of consecutive frames starting at start
          Falls back to one copy if the underlying memoryview is not C-contiguous.
        :param start: logical start index for view
        :param frames: number of frames to show
        """
        start = int(start)
        frames = int(frames)

        def _call_view(start_i, frames_i, dim0=None, dim1=None):
            try:
                return self._ring.view_window(start_i, frames_i)
            except TypeError:
                if dim0 is None or dim1 is None:
                    raise
                return self._ring.view_window(start_i, frames_i, int(dim0), int(dim1))

        if self._mode == "line":
            N, C = self._N, self._C
            mv = _call_view(start, frames, C, N)
            try:
                arr = np.frombuffer(mv, dtype=self.dtype, count=frames * C * N)
            except BufferError:
                arr = np.frombuffer(mv.tobytes(), dtype=self.dtype, count=frames * C * N)
            return arr.reshape(frames, C, N)

        H, W, Cimg = self._H, self._W, self._Cimg
        mv = _call_view(start, frames, H, W * Cimg)

        elem_count = frames * H * W * Cimg
        try:
            arr = np.frombuffer(mv, dtype=self.dtype, count=elem_count)
            arr = arr.reshape(frames, H, W, Cimg)
            return arr
        except BufferError:
            arr = np.frombuffer(mv.tobytes(), dtype=self.dtype, count=elem_count)
            return arr.reshape(frames, H, W, Cimg)

    def view_window_bytes(self, start: int, frames: int):
        if frames <= 0:
            return memoryview(b"")
        if self._mode == 'line':
            return self._ring.view_window(int(start), int(frames), int(self._C), int(self._S))
        else:
            return self._ring.view_window(int(start), int(frames), int(self._H), int(self._W * self._Cimg))


from sqlitedict import SqliteDict
import numpy as np
import pathlib
from typing import *


def create_sqlite3_file(filepath: str, key: str, dtype=np.float32):
    """Create (or ensure) a sqlite3 key."""
    with SqliteDict(filepath) as db:
        if key not in db:
            db[key] = np.array([], dtype=dtype)
        db.commit()

def load_sqlite3_file(filepath: str):
    """Return a live SqliteDict handle (caller manages context)."""
    return SqliteDict(filepath)


class StorageManager:
    def __init__(self, channel_key: Union[np.ndarray, list, tuple, str],
                 filepath: str = './serial_db.sqlite3', overwrite: bool = False):
        """
        :param channel_key: list/array of channel key names (or single key)
        :param filepath: storage file path (.hdf5 or .sqlite3)
        :param overwrite: if True, allow recreating files (hdf5 only here)
        """
        self.channel_key = channel_key if isinstance(channel_key, (list, tuple, np.ndarray)) else [channel_key]
        self.filepath = filepath
        self.filetype = pathlib.Path(self.filepath).suffix
        self.overwrite = overwrite
        if self.filetype not in ".sqlite3":
            raise ValueError(f"defined filetype {self.filetype} is unsupported. Must use .sqlite3")

    def create_serial_database(self, dtype_map: Dict[str, np.dtype] = None):
        """Create a database with the provided channel keys."""
        dtype_map = dtype_map or {}
        with SqliteDict(self.filepath) as db:
            for key in self.channel_key:
                if key not in db:
                    dt = dtype_map.get(key, np.float32)
                    db[key] = np.array([], dtype=dt)
            if 'time' not in db:
                db['time'] = np.array([], dtype=np.float64)
            db.commit()

    @staticmethod
    def load_serial_database(filepath: str = './serial_db.sqlite3', filetype: str = None):
        """Open a DB handle (caller is responsible for closing it if needed)."""
        if filetype is None:
            filetype = pathlib.Path(filepath).suffix

        if filetype == ".sqlite3":
            return load_sqlite3_file(filepath=filepath)
        else:
            raise ValueError(f"filetype {filetype} must be .sqlite3")

    @classmethod
    def load_serial_channel(cls, key: str, filepath: str = './serial_db.sqlite3',
                            filetype: str = None, return_db: bool = False):
        """
        Load a channel w/in sqlite database
        """
        db = cls.load_serial_database(filepath=filepath, filetype=filetype)
        if filetype is None:
            filetype = pathlib.Path(filepath).suffix

        if filetype == ".sqlite3":
            try:
                with db:
                    channel = db[key]
            except Exception:
                raise ValueError(f'Given key {key} is not in sqlite3 file at {filepath}')
            return (db, channel) if return_db else channel
        else:
            raise ValueError(f"filetype {filetype} must be .sqlite3")

    def append_serial_channel(self, key: str, data: np.ndarray):
        """
        Append data to a channel. For SQLite, this **creates the key if missing**.
        """
        if self.filetype == ".sqlite3":
            db = self.load_serial_database(filepath=self.filepath, filetype=self.filetype)
            with db:
                if key not in db:
                    db[key] = np.array([], dtype=data.dtype)
                existing = np.asarray(db[key], dtype=data.dtype)
                db[key] = np.append(existing, data)
                db.commit()
            return
        else:
            raise ValueError(f"Unsupported filetype {self.filetype}")



import numpy as np
import sys
import multiprocessing
from typing import *


def setup_process_start_method():
    """ Set up multiprocessing start method based on operating system

    """
    if sys.platform.startswith('win'):
        multiprocessing.set_start_method("spawn", force=True)
        os_flag = 'win'
    elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin') or sys.platform.startswith('darwin'):
        multiprocessing.set_start_method("fork", force=True)
        os_flag = 'unix'
    else:
        raise EnvironmentError('Unsupported platform')
    return os_flag


def create_static_dict(
    ser_channel_key,
    plot_channel_key,
    commport: str,
    baudrate: int,
    shm_name: str,
    shape: Tuple[int, ...],
    dtype=np.float32,
    ring_capacity: int = 4096,
    *,
    data_mode: str = "line",                 # <— NEW
    frame_shape: Optional[Tuple[int, ...]] = None  # <— NEW (logical shape)
) -> Dict:
    """Build the static args dict that both processes read."""
    return {
        "ser_channel_key": ser_channel_key,
        "plot_channel_key": plot_channel_key,
        "commport": commport,
        "baudrate": baudrate,
        "shm_name": shm_name,
        "shape": tuple(shape),
        "dtype": dtype,
        "ring_capacity": int(ring_capacity),
        "data_mode": str(data_mode),
        "frame_shape": tuple(frame_shape) if frame_shape else None,
    }



def update_static_dict(static_args_dict: dict, **kwargs):
    """ Update existing static args dict
    :param static_args_dict: dictionary to update params
    :param kwargs: all parameters you wish to update
    :return: static_args_dict. Dictionary containing all static parameters
    """
    valid_keys = ['ser_channel_key', 'plot_channel_key',
                  'commport', 'baudrate', 'shm_name',
                  'shape', 'dtype', 'EOL', 'ring_capacity',
                  'num_channel', 'plot_target_fps',
                  'plot_catch_up_max', 'plot_catchup_boost',
                  'plot_lag_frames', 'data_mode']
    for key in kwargs:
        if key in valid_keys:
            static_args_dict[f"{key}"] = kwargs[f"{key}"]
        else:
            Warning(f"Key {key} is not an acceptable input in static_args_dict.\n"
                    f"omitted from dictionary.")
    return static_args_dict


def _coerce(val, default):
    return val if isinstance(val, (int, float)) and not isinstance(val, bool) else default

class DictManager(object):
    def __init__(self, args_dict: dict = None, dict_type: str = None):
        """ Initialize Dictionary Manager class - handles unpacking of argument dicts.

        :param args_dict: argument dictionary. Can be static or dynamic parameter dictionary
        :param dict_type: denotes what type of dictionary to unpack (static or dynamic)
         """
        self.args_dict = args_dict
        self.dict_type = dict_type

    def select_dictionary(self, args_dict: dict, dict_type: str):
        """ Function to update class attribute values for args dict and type
        :param args_dict: argument dictionary. Can be static or dynamic parameter dictionary
        :param dict_type: denotes what type of dictionary to unpack (static or dynamic)
        """
        self.args_dict = args_dict
        self.dict_type = dict_type

    def unpack_selected_dict(self):
        """ Selects which dictionary to unpack based upon input arguments
        Key Arguments: dict_type (static or dynamic)
        """
        self.unpack_online_static_dict()


    def unpack_online_static_dict(self):
        """ Unpack static parameter dictionary for online/real-time use
        :return: adds attributes to self for each key in dict
        """
        essential_keys = ['ser_channel_key', "plot_channel_key", 'commport',
                          'baudrate', 'shm_name', 'shape', 'dtype', 'ring_capacity',
                          'data_mode']
        optional_keys = ['EOL', 'num_points', 'num_channel', 'plot_target_fps',
                         'plot_catchup_base_max', 'plot_catchup_boost']

        for key in essential_keys:
            try:
                setattr(self, f"{key}", self.args_dict[f"{key}"])
            except KeyError:
                raise KeyError(f"selected dictionary doesn't have key {key}\n"
                               f"dictionary must include the following keys:\n"
                               f"{essential_keys}")

        for key in optional_keys:
            try:
                setattr(self, f"{key}", self.args_dict[f"{key}"])
            except KeyError:
                if key == "num_channel":
                    num_channel = np.shape(self.args_dict["plot_channel_key"])[0] * \
                                  np.shape(self.args_dict["plot_channel_key"])[1]
                    setattr(self, f"{key}", num_channel)
                else:
                    setattr(self, f"{key}", None)




import time
import os
from sensor_core.data import DataManager
from sensor_core.memory.strg_manager import StorageManager
from sensor_core.dsp.dsp_manager import DSPManager
from sensor_core.memory.mem_utils import *
from sensor_core.utils.utils import *
from sensor_core.utils.utils import _coerce
from multiprocessing import Process, freeze_support, Manager
from threading import Thread
import pathlib


class SensorManager(DataManager, StorageManager):
    def __init__(self, 
                 ser_channel_key: Union[np.ndarray, str],
                 commport: str,
                 baudrate: int = 115200,
                 dtype=np.float32,
                 data_mode: str = "line",
                 frame_shape: tuple = (1000, 100, 3),
                 fast_stream_path_a: str="./serial_stream_a.bin",
                 fast_stream_path_b: str = "./serial_stream_b.bin",
                 start_stream_ingest: bool = False,
                 sqlite_path: str = "./serial_db.sqlite3",
                 rotate_frames: int = 8192,
                 rotate_seconds: float = 5.0,
                 **kwargs):
        """ Initialize SensorManager Class
        Initializes serial port, shared memory object, and kwarg dictionary (args_dict)
        :param serial_channel_key: list of serial channel names
        :param commport: target serial port
        :param baudrate: target data transfer rate (in bits/sec)
        :param frame_shape: for line data, tuple of (num_points, window_size, num_channels); for image data, tuple of (height, width, num_channels)
        :param dtype: data type to store in shared memory object
        """
        self.dtype = dtype
        self.data_mode = data_mode

        # Defines start method for multiprocessing. Necessary for windows and macOS
        self.os_flag = setup_process_start_method()

        # Setup serial and plot channels
        self.ser_channel_key, self.plot_channel_key = self.setup_channel_keys(
                                                      ser_channel_key=ser_channel_key,
                                                      **kwargs)
        # Setup ring buffer
        self.ring, self.logical_shape = initialize_ring(ser_channel_key=ser_channel_key,
                                                        dtype=dtype,
                                                        data_mode=data_mode,
                                                        frame_shape=frame_shape)
                        
        # Setup target consumer params and enforce
        plot_target_fps = kwargs.get("plot_target_fps", 60.0)
        plot_catch_up_max = kwargs.get("plot_catchup_base_max", 2048)
        plot_catchup_boost = kwargs.get("plot_catchup_boost", 2.5)

        plot_target_fps = _coerce(plot_target_fps, 60.0)
        plot_catch_up_max = int(_coerce(plot_catch_up_max, 2048))
        plot_catchup_boost = _coerce(plot_catchup_boost, 2.5)

        # Setup static args dict
        self.static_args_dict = create_static_dict(ser_channel_key=self.ser_channel_key,
                                                   plot_channel_key=self.plot_channel_key,
                                                   commport=commport,
                                                   baudrate=baudrate,
                                                   shm_name='/sensor_ring',
                                                   shape=self.logical_shape,
                                                   dtype=dtype,
                                                   ring_capacity=4096,
                                                   data_mode=data_mode,
                                                   frame_shape=self.logical_shape
                                                   )
        
        self.static_args_dict = update_static_dict(static_args_dict=self.static_args_dict,
                                                   plot_target_fps=plot_target_fps,
                                                   plot_catch_up_max=plot_catch_up_max,
                                                   plot_catchup_boost=plot_catchup_boost
                                                   )
       

        # Make shared proxies for metrics
        self._mp_manager = Manager()
        self.writer_metrics_proxy = self._mp_manager.dict()
        self.plot_metrics_proxy = self._mp_manager.dict()
        self.ingest_metrics_proxy = self._mp_manager.dict()
        self.stream_ctrl_proxy = self._mp_manager.dict()
        # Initialize proxies
        self.writer_metrics_proxy.update({"init": True})
        self.plot_metrics_proxy.update({"init": True})
        self.ingest_metrics_proxy.update({"init": True})
        self.stream_ctrl_proxy.update({"force_rotate": False})

        # Make shared proxy for DSP
        self._plot_dsp = DSPManager()
        self._plot_dsp_version = 0

        self.plot_dsp_proxy = self._mp_manager.dict()
        self.plot_dsp_proxy.update({
            "version": 0,
            "queue": [],
            "modules": {},
        })

        # Derive default bin paths
        if start_stream_ingest and (fast_stream_path_a is None or fast_stream_path_b is None):
            base = pathlib.Path(sqlite_path).with_suffix('').as_posix()
            fast_stream_path_a = fast_stream_path_a or f"{base}_stream_a.bin"
            fast_stream_path_b = fast_stream_path_b or f"{base}_stream_b.bin"

        # Normalize ALL paths to absolute
        if sqlite_path is not None:
            sqlite_path = os.path.abspath(sqlite_path)
        if fast_stream_path_a is not None:
            fast_stream_path_a = os.path.abspath(fast_stream_path_a)
        if fast_stream_path_b is not None:
            fast_stream_path_b = os.path.abspath(fast_stream_path_b)

        self.fast_stream_path_a = fast_stream_path_a
        self.fast_stream_path_b = fast_stream_path_b
        self.sqlite_path = sqlite_path

        auto_enable_for_image = (self.data_mode.lower() == "image")
        ingest_enabled = bool(start_stream_ingest or auto_enable_for_image)

        # Setup on-disk dual bytestream
        try:
            from sensor_core.memory.stream_logger import dump_loop as _dump_loop
            ring_args = self.static_args_dict
            shm_name = ring_args.get('shm_name', '/sensor_ring')
            capacity = int(ring_args.get('ring_capacity', 4096))
            frame_shape = ring_args.get('logical_shape', self.logical_shape)
            _dtype = ring_args.get("dtype", dtype)
            self._stream_proc = Process(target=_dump_loop,
                                        args=(fast_stream_path_a, fast_stream_path_b, shm_name, capacity, frame_shape, _dtype),
                                        kwargs={'overwrite': False,
                                                'rotate_frames': int(rotate_frames),
                                                'rotate_seconds': float(rotate_seconds) if rotate_seconds else None,
                                                'metrics_proxy': self.writer_metrics_proxy,
                                                'control_proxy': self.stream_ctrl_proxy,
                                                'data_mode': self.data_mode,
                                                })
            self.start_process(self._stream_proc)
        except Exception as e:
            print(f'[SensorManager] failed to start stream logger: {e}')
            self._stream_proc = None

        if start_stream_ingest:
            if ingest_enabled:
                try:
                    from sensor_core.memory.db_ingester import ingest_loop as _ingest_loop
                    ch_keys = list(self.ser_channel_key) if isinstance(self.ser_channel_key, (list, tuple, np.ndarray)) else [self.ser_channel_key]
                    frame_shape = ring_args.get('logical_shape', self.logical_shape)
                    _dtype = ring_args.get("dtype", dtype)
                    # TO DO: change away from 'hint' and just call them the actual kwargs (i.e. frame)shape, data_mode)
                    self._ingest_proc = Process(target=_ingest_loop,
                                                args=(fast_stream_path_a, fast_stream_path_b, sqlite_path, ch_keys),
                                                kwargs={'metrics_proxy': self.ingest_metrics_proxy,
                                                        'data_mode_hint': data_mode,
                                                        'frame_shape_hint': frame_shape,
                                                        'dtype_hint': _dtype,
                                                        'precreate_sqlite': True
                                                        })
                    self.ingest_metrics_proxy.update({
                        "ingest_config_enabled": True,
                        "ingest_config_reason": ("explicit_flag" if start_stream_ingest else "auto_image_mode")
                    })
                    self.start_process(self._ingest_proc)
                except Exception as e:
                    print(f'[SensorManager] failed to start stream ingester: {e}')
                    self.ingest_metrics_proxy.update({
                        "ingest_config_enabled": True,
                        "ingest_last_error": f"spawn_failed: {e.__class__.__name__}: {e}",
                    })
        else:
            self._ingest_proc = None
            self.ingest_metrics_proxy.update({
                "ingest_config_enabled": False,
                "ingest_config_reason": "disabled_by_config"
            })

        if self._ingest_proc is not None:
            t = Thread(target=self._watch_ingester, daemon=True)
            t.start()

    @staticmethod
    def setup_channel_keys(ser_channel_key, **kwargs):
        """ Set up serial and plot channel keys
        :param ser_channel_key: serial channel key (list of names for serial channels)
        :param kwargs: can contain plot_channel_key kwarg, which we can use to set plot settings
        :return: ser_channel_key and plot_channel_key, lists of keys for serial and plot functions
        """
        if len(np.shape(ser_channel_key)) > 1:
            raise ValueError(f"serial channel key {ser_channel_key} \n"
                             f"must be a one-dimensional list")
        if "plot_channel_key" in kwargs:
            plot_channel_key = kwargs["plot_channel_key"]
        else:
            plot_channel_key = [ser_channel_key]

        plot_shape = np.shape(plot_channel_key)

        for key in np.reshape(plot_channel_key, newshape=(1, plot_shape[0] * plot_shape[1]))[0]:
            if key not in ser_channel_key:
                raise KeyError(f'plot_channel_key must include only keys within serial_channel_key')

        return ser_channel_key, plot_channel_key

    def update_data_process(self, save_data: bool = False, filepath: str = None,
                            virtual_ser_port: bool = False, func=None):
        """ Initialize dedicated process to update data

        :param save_data: boolean flag. If true, save acquired data to file and RAM, if not just update it in RAM
        :param filepath: string denoting target filepath to create database
        :param virtual_ser_port: boolean, if true data manager will not instantiate serial port. Will rely on
        user-defined custom function to generate simulated data
        :param func: optional custom function to handle serial data acquisition
        :return: pointer to process
        """
        if save_data and (filepath is not None):
            filetype = pathlib.Path(filepath).suffix
            if filetype != ".hdf5" and filetype != ".sqlite3":
                raise ValueError(f"filepath {filepath} is an invalid filetype {filetype}\n"
                                 f"filepaths should create .hdf5 or .sqlite3 files only")

        odm = DataManager(static_args_dict=self.static_args_dict,
                          save_data=save_data,
                          filepath=filepath,
                          virtual_ser_port=virtual_ser_port,
                          metrics_proxy=self.writer_metrics_proxy)

        if self.os_flag == 'win':
            p = Thread(name='update',
                       target=odm.online_update_data,
                       args=(func,))
        else:
            p = Process(name='update',
                        target=odm.online_update_data,
                        args=(func,))

        return p

    def setup_plotting_process(self):
        """ Initialize dedicated process to update plot
        :return: pointer to process and plot object
        """
        from sensor_core.plot import PlotManager

        pm = PlotManager(static_args_dict=self.static_args_dict,
                         metrics_proxy=self.plot_metrics_proxy,
                         plot_dsp_proxy=self.plot_dsp_proxy,)
        if self.os_flag == 'win':
            p = Thread(name='plot',
                       target=pm.online_plot_data)
        else:
            p = Process(name='plot',
                        target=pm.online_plot_data)

        return p, pm.fig



    def start_process(self, process):
        """ Function to start given process, and ensure safe operability with windows

        :param process: process object to start
        """
        if self.os_flag == 'win':
            freeze_support()
            process.start()
        else:
            process.start()

    def get_metrics(self) -> dict:
        """Return a combined metrics snapshot."""
        return {
            "writer": dict(self.writer_metrics_proxy),
            "plot": dict(self.plot_metrics_proxy),
            "ingest": dict(self.ingest_metrics_proxy),
        }

    def force_seal_now(self):
        """Request the writer to seal current bin and switch immediately."""
        self.stream_ctrl_proxy.update({"force_rotate": True})

    def debug_status(self) -> dict:
        return {
            "writer_proc": {
                "pid": getattr(self._stream_proc, "pid", None),
                "alive": (self._stream_proc.is_alive() if self._stream_proc else False),
            },
            "ingest_proc": {
                "pid": getattr(self._ingest_proc, "pid", None),
                "alive": (self._ingest_proc.is_alive() if self._ingest_proc else False),
            },
            "paths": {
                "bin_a": self.fast_stream_path_a,
                "bin_b": self.fast_stream_path_b,
                "sqlite": self.sqlite_path,
            },
        }

    def _watch_ingester(self):
        while True:
            time.sleep(1.0)
            if not (self._ingest_proc and self._ingest_proc.is_alive()):
                self.ingest_metrics_proxy.update({
                    "ingest_alive": False,
                    "ingest_last_error": self.ingest_metrics_proxy.get("ingest_last_error", "ingester process not alive"),
                    "ingest_updated_unix": time.time(),
                })
                break

    def _publish_plot_dsp_cfg(self):
        self._plot_dsp_version += 1
        self.plot_dsp_proxy["version"] = int(self._plot_dsp_version)
        self.plot_dsp_proxy["queue"] = list(self._plot_dsp.dsp_modules_queue)
        self.plot_dsp_proxy["modules"] = dict(self._plot_dsp.dsp_modules)

    # UI entrypoints
    def add_plot_dsp_module(self, name: str, target_algo: str, **kwargs):
        """
        Add a visualization-only DSP module. Does not affect stored data.
        Example:
          add_plot_dsp_module("ma1", "moving_average_filter", window_size=16, pad="percentile")
          add_plot_dsp_module("bp1", "butterworth_filter", order=2, min_frq=1, max_frq=20, fs=200)
        """
        self._plot_dsp.add_dsp_module(name, target_algo, **kwargs)
        self._publish_plot_dsp_cfg()

    def remove_plot_dsp_module(self, name: str):
        self._plot_dsp.remove_dsp_module(name)
        self._publish_plot_dsp_cfg()

    def clear_plot_dsp_modules(self):
        self._plot_dsp.dsp_modules.clear()
        self._plot_dsp.dsp_modules_queue.clear()
        self._publish_plot_dsp_cfg()

    def set_plot_dsp_order(self, new_queue):
        new_queue = list(new_queue)
        if set(new_queue) != set(self._plot_dsp.dsp_modules_queue):
            raise ValueError("new_queue must contain exactly the existing DSP module names.")
        self._plot_dsp.dsp_modules_queue = new_queue
        self._publish_plot_dsp_cfg()

    def get_plot_dsp_cfg(self) -> dict:
        return {
            "version": int(self._plot_dsp_version),
            "queue": list(self._plot_dsp.dsp_modules_queue),
            "modules": dict(self._plot_dsp.dsp_modules),
        }



__all__ = ["SensorManager"]

def __getattr__(name):
    if name == "SensorManager":
        from .sensor_manager import SensorManager
        return SensorManager
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")




import os, json, struct, time, traceback
from typing import List, Optional, Tuple
import numpy as np
from sqlitedict import SqliteDict
from .strg_manager import StorageManager

MAGIC = b'SCBIN\x00\x00'
MAGIC_LEN = len(MAGIC)
REC_HEADER_SZ = 16

def _seal_path(p: str) -> str: return p + ".seal"

def _read_header(fh):
    magic = fh.read(MAGIC_LEN)
    if magic != MAGIC:
        raise ValueError('Invalid stream magic')
    ver_bytes = fh.read(2);  ver = int.from_bytes(ver_bytes, 'little')
    len_bytes = fh.read(4);  hdr_len = int.from_bytes(len_bytes, 'little')
    payload = fh.read(hdr_len)
    hdr = json.loads(payload.decode('utf-8'))
    return ver, hdr, ver_bytes, len_bytes, payload

def _ensure_sqlite_keys_line(sqlite_path: str, channel_keys: List[str], dtype: np.dtype):
    with SqliteDict(sqlite_path) as db:
        for k in channel_keys:
            if k not in db:
                db[k] = np.array([], dtype=dtype)
        if 'time' not in db:
            db['time'] = np.array([], dtype=np.float64)
        db.commit()

def _ensure_sqlite_keys_image(sqlite_path: str, shape: Tuple[int,int,int], dtype: np.dtype):
    with SqliteDict(sqlite_path) as db:
        if 'image' not in db:
            db['image'] = np.array([], dtype=dtype)  # flattened frames appended
        if 'image_shape' not in db:
            db['image_shape'] = tuple(shape)         # (H,W,Cimg)
        db.commit()

def _ingest_file_line(path: str, sqlite_path: str, channel_keys: List[str],
                      batch_frames: int, dtype: np.dtype, C: int, S: int,
                      metrics_accum: dict):
    _ensure_sqlite_keys_line(sqlite_path, channel_keys, dtype)
    sm = StorageManager(channel_key=channel_keys, filepath=sqlite_path, overwrite=False)
    acc = {k: [] for k in channel_keys}
    frames = 0; bytes_read = 0; batches = 0
    with open(path, 'rb') as fh:
        ver, hdr, ver_b, len_b, payload = _read_header(fh)
        while True:
            rec = fh.read(REC_HEADER_SZ)
            if not rec: break
            ts_ns, wi = struct.unpack('<QQ', rec)
            raw = fh.read(S * C * dtype.itemsize)
            if len(raw) < S * C * dtype.itemsize:
                break
            arr = np.frombuffer(raw, dtype=dtype, count=C*S).reshape(C, S)
            for ci, key in enumerate(channel_keys):
                acc[key].append(arr[ci])
            frames += 1
            bytes_read += REC_HEADER_SZ + len(raw)
            if sum(len(v) for v in acc.values()) >= batch_frames:
                for key in channel_keys:
                    if acc[key]:
                        block = np.concatenate(acc[key], axis=0)
                        sm.append_serial_channel(key, block)
                        acc[key].clear()
                batches += 1
    for key in channel_keys:
        if acc[key]:
            block = np.concatenate(acc[key], axis=0)
            sm.append_serial_channel(key, block)
            acc[key].clear()
    metrics_accum["frames_ingested"] = metrics_accum.get("frames_ingested", 0) + frames
    metrics_accum["bytes_read"] = metrics_accum.get("bytes_read", 0) + bytes_read
    metrics_accum["batches_flushed"] = metrics_accum.get("batches_flushed", 0) + batches
    return ver_b, len_b, payload

def _ingest_file_image(path: str, sqlite_path: str, shape: Tuple[int,int,int],
                       batch_frames: int, dtype: np.dtype, metrics_accum: dict):
    H, W, Cimg = shape
    frame_items = H * W * Cimg
    _ensure_sqlite_keys_image(sqlite_path, shape, dtype)
    sm = StorageManager(channel_key=['image'], filepath=sqlite_path, overwrite=False)
    acc = []
    frames = 0; bytes_read = 0; batches = 0
    with open(path, 'rb') as fh:
        ver, hdr, ver_b, len_b, payload = _read_header(fh)
        while True:
            rec = fh.read(REC_HEADER_SZ)
            if not rec: break
            ts_ns, wi = struct.unpack('<QQ', rec)
            raw = fh.read(frame_items * dtype.itemsize)
            if len(raw) < frame_items * dtype.itemsize:
                break
            arr = np.frombuffer(raw, dtype=dtype, count=frame_items)  # flat
            acc.append(arr)
            frames += 1
            bytes_read += REC_HEADER_SZ + len(raw)
            if len(acc) >= batch_frames:
                block = np.concatenate(acc, axis=0)
                sm.append_serial_channel('image', block)
                acc.clear()
                batches += 1
    if acc:
        block = np.concatenate(acc, axis=0)
        sm.append_serial_channel('image', block)
        acc.clear()
    metrics_accum["frames_ingested"] = metrics_accum.get("frames_ingested", 0) + frames
    metrics_accum["bytes_read"] = metrics_accum.get("bytes_read", 0) + bytes_read
    metrics_accum["batches_flushed"] = metrics_accum.get("batches_flushed", 0) + batches
    return ver_b, len_b, payload

def ingest_loop(file_a: str, file_b: str, sqlite_path: str, channel_keys: List[str],
                batch_frames: int = 32, sleep_s: float = 0.2,
                metrics_proxy: Optional[dict] = None,
                data_mode_hint: Optional[str] = None,
                frame_shape_hint: Optional[Tuple[int, ...]] = None,
                dtype_hint: Optional[str] = None,
                precreate_sqlite: bool = True):

    try:
        # initialize metrics immediately
        if metrics_proxy is not None:
            metrics_proxy.update({
                "ingest_bins_ingested": 0,
                "ingest_frames_ingested": 0,
                "ingest_bytes_read": 0,
                "ingest_batches_flushed": 0,
                "ingest_fps_estimate": 0.0,
                "ingest_updated_unix": time.time(),
                "ingest_alive": True,
                "ingest_watch_paths": [os.path.abspath(file_a), os.path.abspath(file_b)],
                "ingest_sqlite_path": os.path.abspath(sqlite_path),
                "ingest_started": True,
            })
    except Exception:
        pass

    try:
        if metrics_proxy is not None:
            metrics_proxy.update({
                "ingest_alive": True,
                "ingest_starting": False
            })

        def _publish_scan(seal_a: str, seal_b: str):
            if metrics_proxy is None:
                return

            exists_a = os.path.exists(seal_a);
            exists_b = os.path.exists(seal_b)
            mt_a = os.path.getmtime(seal_a) if exists_a else None
            mt_b = os.path.getmtime(seal_b) if exists_b else None
            metrics_proxy.update({
                "ingest_scan_seals": [seal_a, seal_b],
                "ingest_scan_exists": [exists_a, exists_b],
                "ingest_scan_mtime": [mt_a, mt_b],
                "ingest_alive": True,
                "ingest_updated_unix": time.time(),
            })

        if precreate_sqlite:
            try:
                if (data_mode_hint or '').lower() == 'image':
                    if frame_shape_hint is not None and dtype_hint is not None:
                        _ensure_sqlite_keys_image(sqlite_path, tuple(frame_shape_hint), np.dtype(dtype_hint))
                else:
                    _ensure_sqlite_keys_line(sqlite_path, channel_keys, np.dtype(dtype_hint or np.float32))
            except Exception:
                pass

        last_frames_total = 0
        last_t = time.monotonic()

        def _metrics_flush(force=False):
            nonlocal last_frames_total, last_t
            if metrics_proxy is None:
                return
            now = time.monotonic()
            dt = now - last_t
            if force or dt >= 1.0:
                frames_total = int(metrics_proxy.get("ingest_frames_ingested", 0))
                fps = (frames_total - last_frames_total) / max(1e-6, dt)
                metrics_proxy.update({
                    "ingest_fps_estimate": float(fps),
                    "ingest_updated_unix": time.time(),
                    "ingest_alive": True,
                    "ingest_starting": False,
                })
                last_frames_total = frames_total
                last_t = now

        files = [file_a, file_b]
        while True:
            did_work = False
            seal_a = _seal_path(file_a)
            seal_b = _seal_path(file_b)
            _publish_scan(seal_a, seal_b)
            for path in files:
                seal = _seal_path(path)
                if os.path.exists(seal):
                    with open(path, 'rb') as fh:
                        try:
                            ver, hdr, ver_b, len_b, payload = _read_header(fh)
                            if metrics_proxy is not None:
                                metrics_proxy.update({
                                    "ingest_last_header": {
                                        "path": os.path.abspath(path),
                                        "data_mode": hdr.get('data_mode', 'line'),
                                        "frame_shape": tuple(hdr.get('frame_shape', [])),
                                        "dtype": hdr.get('dtype'),
                                    }
                                })
                        except ValueError as e:
                            if metrics_proxy is not None:
                                metrics_proxy.update({
                                    "ingest_last_error": f"HeaderError on {os.path.abspath(path)}: {e}",
                                })
                            continue
                    dtype = np.dtype(hdr['dtype'])
                    shape = tuple(hdr['frame_shape'])
                    mode = hdr.get('data_mode', 'line')

                    delta = {"frames_ingested": 0, "bytes_read": 0, "batches_flushed": 0}
                    if mode == 'line':
                        N, _, C = shape
                        _ = _ingest_file_line(path, sqlite_path, channel_keys, batch_frames, dtype, N, C, delta)
                    elif mode == 'image':
                        H, W, Cimg = shape
                        _ = _ingest_file_image(path, sqlite_path, (H, W, Cimg), batch_frames, dtype, delta)
                    else:
                        continue

                    if metrics_proxy is not None:
                        metrics_proxy.update({
                            "ingest_bins_ingested": int(metrics_proxy["ingest_bins_ingested"]) + 1,
                            "ingest_frames_ingested": int(metrics_proxy["ingest_frames_ingested"]) + int(delta["frames_ingested"]),
                            "ingest_bytes_read": int(metrics_proxy["ingest_bytes_read"]) + int(delta["bytes_read"]),
                            "ingest_batches_flushed": int(metrics_proxy["ingest_batches_flushed"]) + int(delta["batches_flushed"]),
                        })
                        _metrics_flush(force=True)

                    # truncate back to header
                    try: os.remove(seal)
                    except FileNotFoundError: pass
                    with open(path, 'wb') as out:
                        out.write(MAGIC); out.write(ver_b); out.write(len_b); out.write(payload)
                    did_work = True

            if not did_work:
                time.sleep(sleep_s)
            _metrics_flush(force=False)
    except Exception as e:
        if metrics_proxy is not None:
            metrics_proxy.update({
                "ingest_alive": False,
                "ingest_last_error": f"{e.__class__.__name__}: {e}",
                "ingest_last_traceback": ''.join(traceback.format_exc())[-2000:],
                "ingest_updated_unix": time.time(),
                "ingest_starting": False,
            })



from sensor_core.utils.metrics import RingMetrics, timer
from sensor_core.memory.ring_adapter import RingBuffer
from sensor_core.memory.mem_utils import _assert_ring_layout
import time
from .plot_utils import *
from sensor_core.utils.utils import DictManager, _coerce
from sensor_core.memory.strg_manager import StorageManager
from typing import Union
from sensor_core.dsp.dsp_manager import DSPManager


def run_once(f):
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return f(*args, **kwargs)
    wrapper.has_run = False
    return wrapper

class PlotManager(DictManager):
    def __init__(self, static_args_dict, metrics_proxy=None, plot_dsp_proxy=None):
        """ Initialize Plot Manager Class
        :param static_args_dict: dictionary with static parameters
        :param metrics_proxy: shared proxy for metrics to pass to parent class
        """
        DictManager.__init__(self)

        self.select_dictionary(args_dict=static_args_dict, dict_type="static")
        self.unpack_selected_dict()
        # Set consumer params
        self.plot_target_fps = _coerce(getattr(self, "plot_target_fps", None), 60.0)
        self.plot_catchup_base_max = int(_coerce(getattr(self, "plot_catchup_base_max", None), 2048))
        self.plot_catchup_boost = _coerce(getattr(self, "plot_catchup_boost", None), 2.5)

        # DSP Manager setup
        self._plot_dsp_proxy = plot_dsp_proxy
        self._plot_dsp_local = DSPManager()
        self._plot_dsp_version_seen = -1

        # Open ring (consumer)
        self.ring = RingBuffer(self.shm_name, int(self.ring_capacity),
                               tuple(self.shape), self.data_mode, self.dtype, create=False)
        _assert_ring_layout(self.ring, tuple(self.shape), self.dtype)

        self.metrics = RingMetrics()

        self.fig = None
        self.initialize_fig()

        # Initialize staging once
        if self.data_mode == 'line':
            self._init_ring_plotting_line()
        else:
            self._init_ring_plotting_image()

        self._metrics_proxy = metrics_proxy
        self._last_push = 0.0
        self._last_read_idx = None

        # Register animation once (you had this twice)
        self.fig.add_animations(self.online_plot_data)

    def _sync_plot_dsp_if_needed(self):
        if self._plot_dsp_proxy is None:
            return

        try:
            version = int(self._plot_dsp_proxy.get("version", 0))
        except Exception:
            version = 0

        if version == self._plot_dsp_version_seen:
            return

        modules = self._plot_dsp_proxy.get("modules", {})
        queue = self._plot_dsp_proxy.get("queue", [])

        # rebuild local pipeline from plain dict/list
        self._plot_dsp_local.dsp_modules = dict(modules) if modules else {}
        self._plot_dsp_local.dsp_modules_queue = list(queue) if queue else []

        self._plot_dsp_version_seen = version


    def initialize_fig(self):
        if self.data_mode == "line":
            fig = create_fig(plot_channel_key=self.plot_channel_key)
            ys = initialize_fig_data(num_channel=self.shape[2], num_points=self.shape[0])
            for i, subplot in enumerate(fig):
                idx = divmod(i, np.shape(self.plot_channel_key)[1])
                plot_data = ys[i]
                subplot.add_line(data=plot_data, name=self.plot_channel_key[idx[0]][idx[1]], cmap='jet')
            self.fig = fig
        else:
            fig = create_fig(plot_channel_key=self.plot_channel_key)  # 1x1 layout
            H, W, Cimg = (self.shape[0], self.shape[1], self.shape[2] if len(self.shape) == 3 else 1)
            if Cimg == 1:
                imbuf = np.zeros((H, W), dtype=self.dtype)
            else:
                imbuf = np.zeros((H, W, Cimg), dtype=self.dtype)

            for i, subplot in enumerate(fig):
                im = subplot.add_image(data=imbuf, name="image", cmap="gray")
                im.vmax=255
                im.vmin=0
            self.fig = fig
        return self.fig

    def _init_ring_plotting_line(self):
        self._lines = {}
        self._meta  = {}
        ncols = int(np.shape(self.plot_channel_key)[1])
        for i, subplot in enumerate(self.fig):
            r, c = divmod(i, ncols)
            ch_key = self.plot_channel_key[r][c]
            if hasattr(subplot, "graphics") and subplot.graphics:
                line = subplot.graphics[-1]
            else:
                line = None
            if line is None:
                continue
            self._lines[ch_key] = line
            pos0 = line.data.value  # (S,D)
            S, D = pos0.shape
            stage = np.empty((S, D), dtype=np.float32)
            x0 = pos0[:, 0].astype(np.float32, copy=True)
            stage[:, 0] = x0
            stage[:, 1] = pos0[:, 1]
            if D > 2:
                z0 = pos0[:, 2].astype(np.float32, copy=True)
                stage[:, 2] = z0
            else:
                z0 = None
            self._meta[ch_key] = {"S": S, "D": D, "x0": x0, "z0": z0, "stage": stage}

    def _init_ring_plotting_image(self):
        # store handle to the image graphic
        self._image = None
        for i, subplot in enumerate(self.fig):
            if hasattr(subplot, "graphics") and subplot.graphics:
                for g in subplot.graphics:
                    if getattr(g, "kind", "").lower() == "image" or g.__class__.__name__.lower().startswith("image"):
                        self._image = g
                if self._image is None:
                    self._image = subplot.graphics[-1]

    def online_plot_data(self, *_, **__):
        target_fps = float(getattr(self, "plot_target_fps", 60.0))
        min_dt = 1.0/max(1e-6, target_fps)
        now = time.perf_counter()
        last = getattr(self, "_last_present", 0.0)
        if now - last < min_dt:
            return
        tick_start = time.perf_counter()
        with timer(lambda ms: self.metrics.note_plot_tick(ms, write_idx=int(self.ring.write_idx))):
            try:
                wi = int(self.ring.write_idx)
                if wi <= 0:
                    return
                self.metrics.last_write_idx = wi

                lag = int(getattr(self, "plot_lag_frames", 16))
                cap = int(self.ring.capacity)
                end = wi - lag
                if end < 0:
                    return

                if self.data_mode == "line":
                    N, S = int(self.shape[0]), int(self.shape[1])
                    K = int(np.ceil(N / max(1, S)))
                    start = end - K + 1
                    if start < 0:
                        return

                    slot = start % cap
                    first = min(K, cap - slot)
                    win1 = self.ring.view_window(start, first)
                    rest = K - first
                    if rest:
                        win2 = self.ring.view_window(start + first, rest)
                        win = np.concatenate((win1, win2), axis=0)
                    else:
                        win = win1

                    self.metrics.update_drop_estimate(write_idx_now=wi, frames_read_this_tick=K)

                    block = np.concatenate([win[i] for i in range(win.shape[0])], axis=1)  # (C, K*N)
                    yblock = block[:, -N:]
                    yblock = np.require(yblock, dtype=np.float32, requirements=["C"])
                    if not yblock.flags["OWNDATA"]:
                        yblock = yblock.copy()

                    ncols = int(np.shape(self.plot_channel_key)[1])
                    per_tick_gpu_ms = 0.0
                    for i, subplot in enumerate(self.fig):
                        r, c = divmod(i, ncols)
                        ch_key = self.plot_channel_key[r][c]
                        if ch_key not in self._lines:
                            continue
                        line = self._lines[ch_key]
                        meta = self._meta[ch_key]
                        S_line, D = meta["S"], meta["D"]
                        stage = meta["stage"]
                        x0, z0 = meta["x0"], meta["z0"]

                        # Check DSP Pipeline
                        self._sync_plot_dsp_if_needed()
                        y = yblock[i]
                        if self._plot_dsp_local.dsp_modules_queue:
                            y = np.asarray(y, dtype=np.float32)
                            try:
                                y = self._plot_dsp_local.run_dsp_modules(y)
                            except Exception as e:
                                pass

                        # Ensure correct length for the GPU stage buffer
                        if y.shape[0] != S_line:
                            y = y[-S_line:] if y.shape[0] > S_line else np.pad(y, (S_line - y.shape[0], 0))

                        stage[:, 0] = x0
                        stage[:, 1] = y

                        if D > 2 and z0 is not None:
                            stage[:, 2] = z0

                        t0 = time.perf_counter()
                        line.data[:S_line] = stage[:S_line]
                        per_tick_gpu_ms += (time.perf_counter() - t0) * 1000.0

                    self.metrics.add_gpu_upload_ms(per_tick_gpu_ms)
                    self.metrics.last_read_idx = int(start)
                    self.metrics.frames_lag = int(wi - start)

                else:
                    CATCH_MAX = int(getattr(self, "plot_catch_up_max", 8))

                    if self._last_read_idx is None:
                        frames_to_read = 1
                        start = end
                    else:
                        available = max(0, end - self._last_read_idx)
                        frames_to_read = min(available, CATCH_MAX) if available > 0 else 0
                        start = end - frames_to_read + 1 if frames_to_read > 0 else None

                    if not frames_to_read:
                        self._last_present = time.perf_counter()
                        return

                    # handle wrap case
                    slot = start % cap
                    first = min(frames_to_read, cap - slot)
                    win1 = self.ring.view_window(start, first)
                    rest = frames_to_read - first
                    if rest:
                        win2 = self.ring.view_window(start + first, rest)
                        win = np.concatenate((win1, win2), axis=0)
                    else:
                        win = win1

                    self.metrics.update_drop_estimate(write_idx_now=wi, frames_read_this_tick=frames_to_read)
                    latest = win[-1]

                    if latest.ndim == 3 and latest.shape[2] == 1:
                        latest = latest[:, :, 0]
                    latest = np.require(latest, dtype=np.float32, requirements=["C"])
                    if not latest.flags["OWNDATA"]:
                        latest = latest.copy()

                    # upload to the single image graphic
                    t0 = time.perf_counter()
                    self._image.data[...] = latest
                    self.metrics.add_gpu_upload_ms((time.perf_counter() - t0) * 1000.0)

                    self._last_read_idx = end
                    self.metrics.last_read_idx = int(end)
                    self.metrics.frames_lag = int(wi - end)

                # Bookkeeping & metrics proxy (rate-limited to ~2 Hz)
                self._last_seen_wi = wi
                now = time.perf_counter()
                last_push = getattr(self, "_last_push", 0.0)
                if self._metrics_proxy is not None and (now - last_push) > 0.5:
                    self._metrics_proxy.update(self.metrics.snapshot())
                    self._last_push = now

            except Exception as e:
                import traceback, sys
                print("[plot] exception:", e, file=sys.stderr)
                traceback.print_exc()
                return

    @staticmethod
    def offline_initialize_data(filepath: str, plot_channel_key: Union[np.ndarray, str]):
        """ Extract offline sensor data for set of keys
        :param filepath: define path to database to read data from
        :param plot_channel_key: define set of keys in database to plot data
        :return: x and y values
        """
        ys = []
        plot_shape = np.shape(plot_channel_key)
        for key in np.reshape(plot_channel_key, newshape=(1, plot_shape[0] * plot_shape[1]))[0]:
            data = StorageManager.load_serial_channel(key=key, filepath=filepath)
            ys.append(data)

        num_points = len(ys[0])
        xs = [np.linspace(0, num_points - 1, num_points)]
        return xs, ys

    @classmethod
    def offline_plot_data(cls, filepath: str, plot_channel_key: Union[np.ndarray, str] = None):
        """ Initialize plot for offline data
        :param filepath: define path to database to read data from
        :param plot_channel_key: define set of keys in database to plot data
        :return: return plot object
        """
        if plot_channel_key is None:
            database = StorageManager.load_serial_database(filepath=filepath)
            channel_key = []
            with database as db:
                for key in db.keys():
                    channel_key.append(key)
            plot_channel_keys = [channel_key]
        else:
            plot_channel_keys = plot_channel_key

        ys = cls.offline_initialize_data(filepath=filepath, plot_channel_key=plot_channel_keys)
        for i in range(np.shape(plot_channel_keys)[0]*np.shape(plot_channel_keys)[1]):
            if not ys[i][:]:
                ys[i][:] = np.ones(1000) * np.linspace(0, 1, 1000)

        fig = create_fig(plot_channel_key=plot_channel_keys)

        for i, subplot in enumerate(fig):
            idx = divmod(i, np.shape(plot_channel_keys)[1])
            subplot.add_line(data=ys[i], name=plot_channel_keys[idx[0]][idx[1]], cmap='jet')
            subplot.auto_scale(maintain_aspect=False)

        fig.show()

        return fig



from sensor_core.utils.utils import DictManager
import numpy as np
import scipy.signal as signal

class DSPManager:
    def __init__(self):
        """Initialize empty array and create empty dictionary
        :param dsp_modules_queue: Array of dsp_module object names
        :param dsp_modules: master dsp_module dictionary of dsp_modules containing, name, algo, and key parameters
        """
        self.dsp_modules = {}
        self.dsp_modules_queue = []

    def add_dsp_module(self, name: str, target_algo: str, **kwargs):
        ma_keys = {'window_size': int, 'pad': str}
        bw_keys = {'order': int, 'min_frq': int, 'max_frq': int, 'fs': int}

        if name in self.dsp_modules_queue:
            raise ValueError(f"Inputted name already exists: {name}")

        if target_algo == 'moving_average_filter':
            if not set(ma_keys).issubset(kwargs.keys()):
                raise ValueError(f"Missing required parameters: {ma_keys}")
            unexpected_params = set(kwargs.keys()) - set(ma_keys)
            if unexpected_params:
                raise ValueError(f"Invalid moving average input, unexpected parameters: {unexpected_params}")

            # basic validation
            if int(kwargs["window_size"]) < 1:
                raise ValueError("window_size must be >= 1")
            if kwargs["pad"] not in ("min", "percentile"):
                raise ValueError("pad must be 'min' or 'percentile'")

            self.dsp_modules[name] = {'algo': target_algo, **kwargs}
            self.dsp_modules_queue.append(name)
            return

        if target_algo == 'butterworth_filter':
            if not set(bw_keys).issubset(kwargs.keys()):
                raise ValueError(f"Missing required parameters: {bw_keys}")
            unexpected_params = set(kwargs.keys()) - set(bw_keys)
            if unexpected_params:
                raise ValueError(f"Invalid butterworth input, unexpected parameters: {unexpected_params}")

            fs = int(kwargs["fs"])
            min_frq = float(kwargs["min_frq"])
            max_frq = float(kwargs["max_frq"])
            order = int(kwargs["order"])

            if fs <= 0:
                raise ValueError("fs must be > 0")
            nyq = fs / 2.0
            if not (0.0 < min_frq < max_frq < nyq):
                raise ValueError(f"Butterworth requires 0 < min_frq < max_frq < fs/2. Got min={min_frq}, max={max_frq}, fs={fs}")
            if order < 1:
                raise ValueError("order must be >= 1")

            self.dsp_modules[name] = {'algo': target_algo, **kwargs}
            self.dsp_modules_queue.append(name)
            return

        raise ValueError(f"Unsupported module algorithm: {target_algo}")

    def remove_dsp_module(self, name: str):
        if name not in self.dsp_modules_queue:
            raise ValueError(f"Specified module does not exist: {name}")
        self.dsp_modules_queue.remove(name)
        del self.dsp_modules[name]

    def run_dsp_modules(self, data):
        new_data = np.copy(data)
        for name in self.dsp_modules_queue:
            module = self.dsp_modules[name]
            target_algo = module['algo']

            if target_algo == 'moving_average_filter':
                filtered_data = self.moving_average_filter(
                    new_data,
                    int(module['window_size']),
                    str(module['pad'])
                )

            elif target_algo == 'butterworth_filter':
                filtered_data = self.butterworth_filter(
                    new_data,
                    float(module['min_frq']),
                    float(module['max_frq']),
                    int(module['order']),
                    int(module['fs'])
                )
            else:
                raise ValueError(f"Unsupported module algorithm: {target_algo}")

            new_data = filtered_data
        return new_data

    @staticmethod
    def moving_average_filter(data, window_size: int, pad: str):
        if pad == 'min':
            data_arr = np.concatenate(([np.min(data)] * (window_size), data))
        elif pad == 'percentile':
            data_arr = np.concatenate(([np.percentile(data, 10)] * (window_size), data))
        else:
            raise ValueError("Invalid moving_average_filter pad; use 'min' or 'percentile'")

        cumsum = np.cumsum(data_arr)
        filtered = (cumsum[window_size:] - cumsum[:-window_size]) / window_size
        return filtered

    @staticmethod
    def butterworth_filter(data, min_frq: float, max_frq: float, order: int, fs: int):
        b, a = signal.butter(order,
                             [min_frq, max_frq],
                             btype='bandpass',
                             analog=False,
                             output='ba',
                             fs=fs)
        return signal.filtfilt(b, a, data)


import numpy as np
from typing import Tuple, Optional
try:
    from sensor_core.memory.ring_adapter import RingBuffer
except Exception as e:
    RingBuffer = None  # will raise when used

def _normalize_image_shape(shape):
    """
    Normalize image shape for C frames
    :param shape: input image shape
    """
    if len(shape) == 2:
        H, W = shape
        return (int(H), int(W), 1)
    elif len(shape) == 3:
        H, W, C = shape
        return (int(H), int(W), int(C))
    else:
        raise ValueError(f"image frame_shape must be (H,W) or (H,W,C), got {shape}")

def _assert_ring_layout(ring: RingBuffer, logical_shape: Tuple[int, ...], dtype) -> None:
    """
    Check to catch misconfiguration
    :param ring: Ring buffer
    :param logical_shape: input shape
    :param dtype: data type of input
    """
    if ring.dtype != np.dtype(dtype):
        raise TypeError(f"Ring dtype mismatch: ring={ring.dtype}, expected={np.dtype(dtype)}")
    if ring.logical_shape != tuple(logical_shape):
        raise ValueError(f"Ring logical shape mismatch: ring={ring.logical_shape}, expected={logical_shape}")

def initialize_ring(ser_channel_key, dtype, shm_name="/sensor_ring",
                    frames_capacity=4096, data_mode="line", frame_shape=None):
    """
    Initialize Ring Buffer
    :param ser_channel_key: serial channel key names
    :param window_size: number of frames per acquisition
    :param dtype: data type
    :param shm_name: name of ring buffer object
    :param frames_capacity: max frame capacity of ring buffer
    :param data_mode: allow for line or image data
    :param frame_shape: input shape of each frame
    """
    if frame_shape is None:
        raise ValueError("frame_shape is required")
    if data_mode.lower() == "line":
        logical = (
            int(frame_shape[0]), # number of points,
            int(frame_shape[1]), # window size (number of frames)
            len(ser_channel_key) # number of serial channels
        )
    else:
        logical = _normalize_image_shape(tuple(frame_shape))

    ring = RingBuffer(shm_name, int(frames_capacity), tuple(logical), data_mode, np.dtype(dtype), create=True)
    return ring, logical



import os, json, struct, time, traceback
from typing import Tuple, Optional
import numpy as np
from .ring_adapter import RingBuffer

MAGIC = b'SCBIN\x00\x00'
VERSION = 1

def _ensure_parent(path: str):
    parent = os.path.dirname(os.path.abspath(path))
    if parent:
        os.makedirs(parent, exist_ok=True)

def _seal_path(path: str) -> str:
    return path + ".seal"

def _contiguous_bytes_view(mv: memoryview) -> memoryview:
    try:
        b = mv.cast('B')
        is_c = getattr(b, 'c_contiguous', None)
        if is_c is None:
            is_c = getattr(b, 'contiguous', True)
        if is_c:
            return b
    except TypeError:
        pass
    return memoryview(bytes(mv))

class BinaryStreamWriter:
    def __init__(self, file_a: str, file_b: str, ring_name: str, capacity_frames: int,
                 frame_shape: Tuple[int, ...], dtype, data_mode: str = 'line',
                 rotate_frames: int = 8192, rotate_seconds: Optional[float] = None,
                 overwrite: bool = False, metrics_proxy: Optional[dict] = None,
                 control_proxy: Optional[dict] = None):
        """
        Append-only binary logger to two alternating files with seal markers
        :param file_a: location of .bin file a
        :param file_b: location of .bin file b
        :param ring_name: location of ring buffer
        :param capacity_frames: max capacity of .bin files
        :param frame_shape: logical shape of frame
        :param dtype: data type
        :param data_mode: line or image data
        :param rotate_frames: how many frames to save before rotating to alternate .bin file
        :param rotate_seconds: how many seconds to wait before forcing switch to alternate .bin file
        :param overwrite: flag to overwrite existing bin and sqlite file
        :param metrics_proxy: metrics proxy for timing analysis
        :param control_proxy: contains flag to force switch between .bin files
        """
        self.files = [file_a, file_b]
        self.ring_name = ring_name
        self.capacity_frames = int(capacity_frames)
        self.frame_shape = tuple(frame_shape)
        self.dtype = np.dtype(dtype)
        self.data_mode = data_mode
        self.rotate_frames = int(rotate_frames)
        self.rotate_seconds = float(rotate_seconds) if rotate_seconds else None
        self._active = 0
        self._frames_written_in_active = 0
        self._fh = None

        # proxies
        self._metrics = metrics_proxy
        self._control = control_proxy

        # writer counters
        self._m_total_frames = 0
        self._m_total_bytes = 0
        self._m_rotations = 0

        # timers
        self._m_last_flush = time.monotonic()
        self._m_frames_since = 0
        self._last_rotation_wall = time.time()
        self._last_heartbeat = 0.0

        # Setup bin files
        _ensure_parent(file_a); _ensure_parent(file_b)
        if overwrite:
            for f in self.files:
                try: os.remove(_seal_path(f))
                except FileNotFoundError: pass
                with open(f, 'wb') as fh: self._write_header(fh)

        for f in self.files:
            if not os.path.exists(f) or os.path.getsize(f) == 0:
                with open(f, 'wb') as fh: self._write_header(fh)

        if os.path.exists(_seal_path(self.files[self._active])):
            self._active = 1 - self._active
        if os.path.exists(_seal_path(self.files[self._active])):
            os.remove(_seal_path(self.files[self._active]))

        self._fh = open(self.files[self._active], 'ab')
        self._frames_written_in_active = 0
        self._publish_heartbeat(force=True)

    def _write_header(self, fh):
        header = {
            'ring_name': self.ring_name,
            'frame_shape': self.frame_shape,
            'dtype': str(self.dtype),
            'data_mode': self.data_mode,
            'version': VERSION,
        }
        payload = json.dumps(header).encode('utf-8')
        fh.write(MAGIC)
        fh.write(struct.pack('<H', VERSION))
        fh.write(struct.pack('<I', len(payload)))
        fh.write(payload)
        fh.flush()
        os.fsync(fh.fileno())

    def _publish_heartbeat(self, force=False):
        if self._metrics is None:
            return
        now = time.time()
        if force or (now - self._last_heartbeat) >= 1.0:
            seals = [os.path.abspath(self.files[0]) + ".seal", os.path.abspath(self.files[1]) + ".seal"]
            seal_exists = [os.path.exists(seals[0]), os.path.exists(seals[1])]
            seal_mtime = [
                (os.path.getmtime(seals[0]) if seal_exists[0] else None),
                (os.path.getmtime(seals[1]) if seal_exists[1] else None),
            ]
            dt = max(1e-6, time.monotonic() - self._m_last_flush)
            fps = self._m_frames_since / dt
            self._metrics.update({
                "writer_active_bin": os.path.abspath(self.files[self._active]),
                "writer_total_frames": int(self._m_total_frames),
                "writer_total_bytes": int(self._m_total_bytes),
                "writer_rotations": int(self._m_rotations),
                "writer_fps_estimate": float(fps),
                "writer_last_rotation_unix": self._last_rotation_wall,
                "writer_updated_unix": now,
                "writer_alive": True,
                "writer_watch_seals": seals,
                "writer_seal_exists": seal_exists,
                "writer_seal_mtime": seal_mtime,
            })
            self._m_last_flush = time.monotonic()
            self._m_frames_since = 0
            self._last_heartbeat = now

    def _rotate(self):
        self._fh.flush(); os.fsync(self._fh.fileno()); self._fh.close()
        open(_seal_path(self.files[self._active]), 'wb').close()
        self._active = 1 - self._active
        try: os.remove(_seal_path(self.files[self._active]))
        except FileNotFoundError: pass
        with open(self.files[self._active], 'wb') as fh: self._write_header(fh)
        self._fh = open(self.files[self._active], 'ab')
        self._frames_written_in_active = 0
        self._m_rotations += 1
        self._last_rotation_wall = time.time()
        self._publish_heartbeat(force=True)

    def _maybe_time_rotate(self):
        if self.rotate_seconds is None:
            return
        if (time.time() - self._last_rotation_wall) >= self.rotate_seconds:
            self._fh.flush()
            self._rotate()

    def _maybe_force_rotate(self):
        if self._control is None:
            return
        if bool(self._control.get("force_rotate", False)):
            self._control.update({"force_rotate": False})
            self._fh.flush()
            self._rotate()

    def write_frames(self, buf: memoryview, frame_bytes: int, start_idx: int, nframes: int, ts_ns: int):
        if nframes <= 0:
            self._maybe_time_rotate()
            self._maybe_force_rotate()
            self._publish_heartbeat(force=False)
            return

        b = _contiguous_bytes_view(memoryview(buf))
        remaining = nframes
        idx = 0
        while remaining > 0:
            can_write = min(remaining, max(1, self.rotate_frames - self._frames_written_in_active))
            for i in range(can_write):
                off = (idx + i) * frame_bytes
                self._fh.write(struct.pack('<QQ', ts_ns, (start_idx + idx + i)))
                self._fh.write(b[off:off+frame_bytes])
            self._frames_written_in_active += can_write
            self._m_total_frames += can_write
            self._m_total_bytes += (can_write * (frame_bytes + 16))
            self._m_frames_since += can_write

            idx += can_write
            remaining -= can_write

            if self._frames_written_in_active >= self.rotate_frames:
                self._fh.flush()
                self._rotate()

        self._maybe_time_rotate()
        self._maybe_force_rotate()
        self._publish_heartbeat(force=False)

def dump_loop(file_a: str, file_b: str, shm_name: str, capacity_frames: int,
              frame_shape: Tuple[int, ...], dtype, data_mode: str = 'line',
              poll_hz: float = 400.0, overwrite: bool = False, rotate_frames: int = 8192,
              rotate_seconds: Optional[float] = None, metrics_proxy: Optional[dict] = None,
              control_proxy: Optional[dict] = None):
    if metrics_proxy is not None:
        metrics_proxy.update({
            "writer_alive": True,
            "writer_start_unix": time.time(),
            "writer_watch_bins": [os.path.abspath(file_a), os.path.abspath(file_b)],
            "writer_watch_seals": [os.path.abspath(file_a) + ".seal", os.path.abspath(file_b) + ".seal"],
            "writer_ring": {"name": shm_name, "capacity": int(capacity_frames), "shape": tuple(frame_shape), "dtype": str(np.dtype(dtype))},
        })
    try:
        ring = RingBuffer(shm_name, capacity_frames, frame_shape, data_mode, dtype, create=False)
        writer = BinaryStreamWriter(file_a, file_b, shm_name, capacity_frames, frame_shape, dtype,
                                    data_mode=data_mode, rotate_frames=rotate_frames,
                                    rotate_seconds=rotate_seconds, overwrite=overwrite,
                                    metrics_proxy=metrics_proxy, control_proxy=control_proxy)
        last_idx = int(ring.write_idx)
        frame_bytes = ring.frame_bytes
        period = 1.0 / poll_hz
        cap = int(capacity_frames)

        def _write_window_wrapped(start: int, n: int, ts_ns: int):
            if n <= 0:
                return
            end = (start + n)
            if end <= cap:
                buf = ring.view_window_bytes(start, n)
                writer.write_frames(buf, frame_bytes, start, n, ts_ns)
            else:
                first = cap - start
                second = end - cap
                if first > 0:
                    buf1 = ring.view_window_bytes(start, first)
                    writer.write_frames(buf1, frame_bytes, start, first, ts_ns)
                if second > 0:
                    buf2 = ring.view_window_bytes(0, second)
                    writer.write_frames(buf2, frame_bytes, 0, second, ts_ns)

        while True:
            wi = int(ring.write_idx)
            if wi != last_idx:
                n = (wi - last_idx) % cap
                if n > 0:
                    start = (wi - n) % cap
                    ts_ns = time.time_ns()
                    _write_window_wrapped(start, n, ts_ns)
                    last_idx = wi
            else:
                writer.write_frames(memoryview(b""), frame_bytes, 0, 0, time.time_ns())
            time.sleep(period)
    except Exception as e:
        if metrics_proxy is not None:
            metrics_proxy.update({
                "writer_alive": False,
                "writer_last_error": f"{e.__class__.__name__}: {e}",
                "writer_last_traceback": ''.join(traceback.format_exc())[-2000:],  # tail to keep small
                "writer_updated_unix": time.time(),
            })



import serial
import sys
import glob
import numpy as np
from typing import *
from itertools import product


def find_serial():
    """ Lists serial port names

        :raises EnvironmentError:
            On unsupported or unknown platforms
        :returns:
            A list of the serial ports available on the system
    """
    if sys.platform.startswith('win'):
        ports = ['COM%s' % (i + 1) for i in range(256)]
    elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
        # this excludes your current terminal "/dev/tty"
        ports = glob.glob('/dev/tty[A-Za-z]*')
    elif sys.platform.startswith('darwin'):
        ports = glob.glob('/dev/tty.*')
    else:
        raise EnvironmentError('Unsupported platform')

    result = []
    for port in ports:
        try:
            s = serial.Serial(port)
            s.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass
    return result


class SerialManager:
    def __init__(self, commport: str, baudrate: int, frame_shape: Tuple[int, ...],
                 EOL: str = None, virtual_ser_port: bool = False):
        """ Initialize SerialManager class - manages functions related to instantiating and using serial port

        :param commport: target serial port
        :param baudrate: target baudrate
        :param frame_shape: shape of output data (number of distinct channels, number of frames)
        :param EOL: optional; end of line phrase used to separate timepoints
        :param virtual_ser_port: boolean, if True will not initialize serial port, instead will rely on user-defined
        custom function to generate simulated data
        """
        self.commport = commport
        self.baudrate = baudrate
        self.frame_shape = frame_shape
        self.EOL = EOL
        self.ser = None
        self.virtual_ser_port = virtual_ser_port

    def setup_serial(self):
        """ Sets up given serial port for a given baudrate

        :return: serial object
        """
        if self.virtual_ser_port:
            pass
        else:
            try:
                self.ser = serial.Serial(self.commport, self.baudrate, timeout=0.1)
                return self.ser
            except (OSError, serial.SerialException):
                raise OSError("Error setting up serial port")

    def _acquire_data(self, frame_shape, data_mode):
        """Default (fallback) reader; replace with real protocol."""
        if data_mode == "line":
            N = int(frame_shape[0])  # num_points
            C = int(frame_shape[2])  # channels
            return np.zeros((N, C), dtype=np.float32)
        else:
            H, W, Cimg = (frame_shape[0], frame_shape[1], frame_shape[2] if len(frame_shape) == 3 else 1)
            return np.zeros((H, W, Cimg), dtype=np.float32)

    def acquire_data(self, func=None, data_mode: str='line'):
        """
        Acquire serial data
        :param func: custom acquisition function
        :param data_mode: Line or Image data acquisition
        Returns:
          LINE mode:  (S, C) float32
          IMAGE mode: (H, W, C) uint8/float32
        """
        if func is None:
            channel_data = self._acquire_data(frame_shape=self.frame_shape,
                                              data_mode=data_mode)  # your legacy path for line
        else:
            try:
                # supply expected frame shape to custom func
                expected = getattr(self, "frame_shape", (self.frame_shape[0], self.frame_shape[2]))
                channel_data = func(
                    ser=self.ser,
                    frame_shape=expected,
                )
            except Exception:
                raise ValueError(
                    f'custom function {func} must accept: ser, frame_shape'
                )

        if channel_data is None:
            return None

        arr = np.asarray(channel_data)

        # Line mode
        if data_mode == "line":
            # normalize to (S, C)
            if arr.ndim != 2:
                raise ValueError(f"LINE mode expects 2D (S, C), got {arr.shape}")
            S, C = arr.shape
            if C != self.num_channel and S == self.num_channel:
                arr = arr.T
            if arr.shape != (self.frame_shape[1], self.num_channel):
                raise ValueError(f"LINE data shape {arr.shape} != ({self.frame_shape[1]}, {self.num_channel})")
            return arr.astype(np.float32, copy=False)
        else: # Image mode
            if arr.ndim == 2:
                arr = arr[:, :, None]
            Hexp, Wexp, Cexp = self.frame_shape
            if arr.shape != (Hexp, Wexp, Cexp):
                raise ValueError(f"IMAGE data shape {arr.shape} != expected {(Hexp, Wexp, Cexp)}")
            return arr


from collections import deque
from contextlib import contextmanager
import time
import numpy as np

# Helper functions for mean and 95% percentile
def _avg(seq):  return float(np.mean(seq)) if seq else 0.0
def _p95(seq):  return float(np.percentile(seq, 95)) if seq else 0.0

class RingMetrics:
    """
    Keeps rolling timing metrics and simple rates for producer (writer) and consumer (plot).
    All durations are in milliseconds.
    """
    def __init__(self, window: int = 500):
        """ Initialize Ring Metrics Class

        :param window: number of samples used to calculate metric
        """
        self.publish_ms = deque(maxlen=window)
        self.plot_ms    = deque(maxlen=window)
        self.gpu_ms     = deque(maxlen=window)
        self.acquire_ms = deque(maxlen=window)

        self._last_pub_t  = None
        self._last_plot_t = None
        self.producer_fps = 0.0
        self.consumer_fps = 0.0

        # ring-related live counters
        self.last_write_idx = 0
        self.last_read_idx  = 0
        self.frames_lag     = 0
        self.drops_est      = 0

        # for crude drop estimation across ticks
        self._prev_write_idx = None

    def note_publish(self, ms: float, write_idx: int | None = None):
        self.publish_ms.append(ms)
        now = time.perf_counter()
        if self._last_pub_t is not None:
            dt = now - self._last_pub_t
            if dt > 0:
                # EWMA for stability
                self.producer_fps = 0.9 * self.producer_fps + 0.1 * (1.0 / dt)
        self._last_pub_t = now
        if write_idx is not None:
            self.last_write_idx = int(write_idx)

    def note_plot_tick(self, ms: float, write_idx: int | None = None):
        self.plot_ms.append(ms)
        now = time.perf_counter()
        if self._last_plot_t is not None:
            dt = now - self._last_plot_t
            if dt > 0:
                self.consumer_fps = 0.9 * self.consumer_fps + 0.1 * (1.0 / dt)
        self._last_plot_t = now
        if write_idx is not None:
            self.last_write_idx = int(write_idx)

    def add_gpu_upload_ms(self, ms: float):
        self.gpu_ms.append(ms)

    def add_acquire_ms(self, ms: float):
        self.acquire_ms.append(ms)

    def update_drop_estimate(self, write_idx_now: int, frames_read_this_tick: int):
        # If writer advanced by more than we consumed, the excess are "drops" at this visualization rate.
        if self._prev_write_idx is not None:
            advanced = max(0, int(write_idx_now) - int(self._prev_write_idx))
            excess   = max(0, advanced - frames_read_this_tick)
            self.drops_est += excess
        self._prev_write_idx = int(write_idx_now)

    def snapshot(self) -> dict:
        return dict(
            producer_fps     = round(self.producer_fps, 2),
            consumer_fps     = round(self.consumer_fps, 2),
            publish_avg_ms   = round(_avg(self.publish_ms), 3),
            publish_p95_ms   = round(_p95(self.publish_ms), 3),
            plot_tick_avg_ms = round(_avg(self.plot_ms), 3),
            plot_tick_p95_ms = round(_p95(self.plot_ms), 3),
            gpu_upload_avg_ms= round(_avg(self.gpu_ms), 3),
            gpu_upload_p95_ms= round(_p95(self.gpu_ms), 3),
            acquire_avg_ms   = round(_avg(self.acquire_ms), 3),
            acquire_p95_ms   = round(_p95(self.acquire_ms), 3),
            write_idx        = int(self.last_write_idx),
            read_idx         = int(self.last_read_idx),
            frames_lag       = int(self.frames_lag),
            drops_est        = int(self.drops_est),
        )

@contextmanager
def timer(cb):
    """
    Usage:
        with timer(lambda ms: metrics.note_publish(ms)):
            do_work()
    """
    t0 = time.perf_counter()
    try:
        yield
    finally:
        cb((time.perf_counter() - t0) * 1000.0)



from setuptools import setup, find_packages
from pybind11.setup_helpers import Pybind11Extension, build_ext
import numpy as np
import sys

install_requires = [
    "numpy",
    "pygfx>=0.1.13",
    "jupyterlab",
    "pyserial",
    "fastplotlib",
    "sqlitedict",
    "pybind11",
]

if sys.platform == "win32":
    extra_compile_args = ["/O2"]
else:
    extra_compile_args = ["-O3", "-fvisibility=hidden", "-fPIC"]

ext_modules = [
    Pybind11Extension(
        "fastring",
        sources=[
            "sensor_core/native/fastring/py_module.cpp",
        ],
        include_dirs=[
            "sensor_core/native/fastring",
            np.get_include(),
        ],
        cxx_std=17,
        define_macros=[("PYBIND11_DETAILED_ERROR_MESSAGES", "1")],
        extra_compile_args=extra_compile_args,
    ),
]

setup(
    name="sensor_core",
    version="1.0.0",
    packages=find_packages(),
    install_requires=install_requires,
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
    description="Core API for plotting sensor data in real-time",
)



import numpy as np
from typing import *

def _fastplotlib():
    try:
        import fastplotlib as fpl
        return fpl
    except ImportError as e:
        raise ImportError(
            "Plotting backend could not be initialized. "
            "If you are running headless, install a supported backend (e.g., `pip install glfw` "
            "or `pip install pyside6`) and ensure a display is available."
        ) from e

def create_fig(plot_channel_key: Union[np.ndarray, str]):
    """ Create fastplotlib Figure (collection of subplot(s))

    :param plot_channel_key: names of subplot(s)
    :return: GridPlot object
    """
    fpl = _fastplotlib()
    grid_shape = np.shape(plot_channel_key)

    fig = fpl.Figure(
        shape=grid_shape,
        names=plot_channel_key
    )
    return fig


def initialize_fig_data(num_channel: int, num_points: int):
    """ Initialize GridPlot data

    :param num_channel: number of distinct channels
    :param num_points: number of 'time' points [num_points = time(s) * Hz]
    :return: x and y arrays [y array has shape (num_channel, num_points)]
    """
    ys = np.ones((num_channel, num_points)) * np.linspace(0, 1, num_points)
    return ys



from sensor_core.utils.metrics import RingMetrics, timer
from sensor_core.memory.ring_adapter import RingBuffer
import numpy as np
from sensor_core.memory.mem_utils import _assert_ring_layout
from sensor_core.serial import SerialManager
from sensor_core.utils import DictManager
from sensor_core.memory.strg_manager import StorageManager
from time import perf_counter
import time, traceback


class DataManager(SerialManager, DictManager, StorageManager):
    def __init__(self,
                 static_args_dict: dict, 
                 virtual_ser_port: bool = False,
                 save_data: bool = False, 
                 filepath: str = None, 
                 overwrite_data: bool = True, 
                 metrics_proxy=None):
        """ Online Data Manager
        - handles serial port initialization and management of data for the online (real-time) use case ONL
        :param static_args_dict: dictionary containing key parameters for initialization
        :param virtual_ser_port: boolean, if True will not initialize serial port, instead will rely on user-defined
        custom function to generate simulated data
        :param save_data: boolean, determines whether to save data
        :param filepath: filepath to save data to
        :param overwrite_data: boolean, decides whether to overwrite existing saved data
        """
        self.metrics = RingMetrics()
        self.static_args_dict = static_args_dict
        self.save_data = save_data
        self._metrics_proxy = metrics_proxy

        # Initialize DictManager Subclass
        DictManager.__init__(self)

        # Unpack static_args_dict
        self.select_dictionary(args_dict=self.static_args_dict,
                               dict_type="static")

        self.unpack_selected_dict()

        # Initialize Buffer
        self.ring = RingBuffer(self.shm_name,
                               int(self.ring_capacity), 
                               tuple(self.shape), 
                               self.data_mode,
                               self.dtype,create=False)
        _assert_ring_layout(self.ring, tuple(self.shape), self.dtype)

        # Start serial port
        self.start_serial(virtual_ser_port=virtual_ser_port)

        # Create serial database
        if save_data:
            StorageManager.__init__(self, channel_key=self.ser_channel_key,
                                    filepath=filepath, overwrite=overwrite_data)
            self.create_serial_database()

    def start_serial(self, virtual_ser_port):
        """ Initialize SerialManager subclass, and setup serial port

        """
        SerialManager.__init__(self, commport=self.commport,
                               baudrate=self.baudrate,
                               frame_shape=self.shape,
                               EOL=self.EOL,
                               virtual_ser_port=virtual_ser_port)
        self.setup_serial()

    def online_update_data(self, func=None):
        last_push = perf_counter()
        last_log = time.time()
        while True:
            try:
                with timer(lambda ms: self.metrics.add_acquire_ms(ms)):
                    ys = self.acquire_data(func=func,
                                           data_mode=self.data_mode)
                if ys is None:
                    if time.time() - last_log > 1.0:
                        print("[writer] acquire_data -> None")
                        last_log = time.time()
                    continue

                if self.data_mode=='line':
                    arr = np.asarray(ys)
                    if arr.ndim != 2:
                        raise ValueError(f"[writer] ys ndim={arr.ndim}, expected 2 (N,C), got {arr.shape}")

                    N_in, C_in = arr.shape
                    N_ring, _, C_ring, = self.frame_shape  # ring frame is (N, C)

                    if C_in != C_ring:
                        raise ValueError(f"[writer] channels mismatch: ys (N,{C_in}), ring expects C={C_ring}")

                    # Clamp to ring N
                    N = min(N_in, N_ring)
                    if N != N_ring:
                        arr = arr[:N, :]

                    # Confirm frame is contiguous and transpose, then publish to ring
                    frame = np.ascontiguousarray(arr, dtype=self.dtype)
                    with timer(lambda ms: self.metrics.note_publish(ms, write_idx=int(self.ring.write_idx))):
                        self.ring.publish(frame)
                else:
                    with timer(lambda ms: self.metrics.note_publish(ms)):
                        self.ring.publish(np.asarray(ys, dtype=self.dtype))

                wi = int(self.ring.write_idx)
                self.metrics.last_write_idx = wi

                now = perf_counter()
                if self._metrics_proxy is not None and (now - last_push) > 0.5:
                    self._metrics_proxy.update(self.metrics.snapshot())
                    last_push = now

            except Exception as e:
                print("[writer] EXCEPTION:", repr(e))
                traceback.print_exc()
                time.sleep(0.02)


